/* Encyklopedie jazyka C
 *
 * Kapitola Soubory
 *
 */

/* Copyright (c) David Martinek, 2023
 * david.martinek.78 (zavináč) gmail.com
 *
 * This work is licensed under a Creative Commons
 * Attribution-NonCommercial-ShareAlike 4.0 International License.
 * (CC BY-NC-SA 4.0)
 * https://creativecommons.org/licenses/by-nc-sa/4.0/
 *
 * Toto dílo je licencováno pole licence Creative Commons
 * Uveďte původ-Neužívejte dílo komerčně-Zachovejte licenci 4.0 Mezinárodní
 * (CC BY-NC-SA 4.0)
 * https://creativecommons.org/licenses/by-nc-sa/4.0/deed.cs
 */

#define VERZE "2.1"
#define NAZEV "Soubory"

// TODO Prostuduj si obsah tohoto hlavičkového souboru! Jsou tam zaklínadla,
// bez kterých některé věci ve Windows nefungují správně!
// - barevný výstup
// - printg - český tisk (překódovává z UTF8 na CP852)
// - printh - český tisk + vyprávěcí mód
// - formátovací a formátové značky pro printf/scanf podle normy
// - obsluha nabídek v kapitolách
#include "pomocny-modul.h"

// Implementace nabídkového menu
#include "menu.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <ctype.h>
// #include <math.h>
// #include <float.h>
// #include <limits.h>

#include <errno.h>
#include <locale.h>
#include <wchar.h>
#include <stdint.h>

// Hlavičky funkcí realizující jednotlivé kapitoly.

void souborVJazyceC(void);
void datovyProud(void);
void pripravaSouboru(void);
void testovaniAZpracovaniChyb(void);

void zapisDoSouboru(void);
void cteniZeSouboru(void);
void souboryProCteniIZapis(void);

void binarniSoubory(void);
void binarniStruktury(void);


int main(void)
{
#ifdef __unix__
  setlocale(LC_ALL, "cs_CZ.UTF-8"); // Když je zapnutá lokalizace, perror tiskne hlášení česky.
#else
  setlocale(LC_ALL, "Czech"); // Když je zapnutá lokalizace, perror tiskne hlášení česky.
#endif
  setlocale(LC_NUMERIC, "C"); // Desetinná tečka

  povolBarevnyVystup();
//  prepniRychlostVypraveni(2);

  Tmenu * menu = menuVytvor(
    NADPIS(
      "==============================================================\n"
      "Encyklopedie jazyka C - " NAZEV " (verze " VERZE ")\n"
      "(c) David Martinek, 2023\n"
      "Zároveň s programem sleduj i zdrojový text!\n"
      "==============================================================\n"
    )
    , zobrazMod
  );

  if (menu == NULL)
  {
    printg("Není dost paměti ani pro výrobu menu. Končím program.");
    return 0;
  }

  menuPridej(menu,   0, "\n" SEKCE("Soubory obecně") , NULL);
  menuPridej(menu, 'A', "Soubor v jazyce C\n", souborVJazyceC);
  menuPridej(menu, 'B', "Datový proud - stream\n", datovyProud);
  menuPridej(menu, 'C', "Příprava souboru pro práci\n", pripravaSouboru);
  menuPridej(menu, 'D', "Testování a zpracování chyb\n", testovaniAZpracovaniChyb);

  menuPridej(menu,   0, "\n" SEKCE("Textové soubory") , NULL);
  menuPridej(menu, 'E', "Zápis do souborů a proudů\n", zapisDoSouboru);
  menuPridej(menu, 'F', "Čtení ze souborů a proudů\n", cteniZeSouboru);
  menuPridej(menu, 'G', "Soubory pro čtení i zápis\n", souboryProCteniIZapis);

  menuPridej(menu,   0, "\n" SEKCE("Binární soubory - pro pokročilé (masochisty)") , NULL);
  menuPridej(menu, 'H', "Zpracování binárních souborů\n", binarniSoubory);
  menuPridej(menu, 'I', "Struktury v binárních souborech\n", binarniStruktury);

//  menuPridej(menu, 'J', "\n", test);
//  menuPridej(menu, 'K', "\n", test);
//  menuPridej(menu, 'L', "\n", test);

  menuPridej(menu,   0, "\n" SEKCE("Nastavení programu"), NULL);
  menuPridej(menu, '0', "Nastav vyprávěcí mód\n", nastavRychlostVypraveni);

  while (menuVolba(menu) != 'Q') {}

  printg("Končím program.\n");
  menuZrus(menu);
  return 0;
}


////////////////////////////////////////////////////////////////////////////////
void souborVJazyceC(void)
{
  zacatekKapitoly("Soubor v jazyce C");

  printh(
    "- Soubor jsou data uložená na paměťovém médiu mimo operační paměť.\n"
    "- O soubory se stará ovladač souborového systému, tedy operační systém.\n"
    "- Práce se soubory vyžaduje použití služeb operačního systému, protože\n"
    "  - existují různé souborové systémy (řeší různý HW i formáty),\n"
    "  - některé počítače vůbec soubory nemají (IOT, jednočipy, signálové procesory).\n"
    "\n"
    "- Z pohledu programovacího jazyka je soubor sekvenční posloupnost bajtů.\n"
    "- Různé formáty souborů navrhují sami tvůrci softwaru.\n"
    "\n"
  );

  if (navratNeboSekce("Datový typ soubor v jazyce C")) return;
  if (pokracovat())
  {
    //FILE * soubor; // zatím neinicializovaný ukazatel...

    printh(
      "\n"
      KOD(
      "  FILE * soubor; // zatím neinicializovaný ukazatel...\n"
      )
      "\n"
      "- Pro práci se soubory v jazyce C se používá typ " KK("FILE*") ".\n"
      "- Je to ukazatel na strukturu popisující soubor a jeho aktuální stav.\n"
      "- Nejde použít samotný typ " KK("FILE") ".\n"
      "  - Je to ukrytá struktura - viz kapitola Struktury.\n"
      // Důvod, proč jsou položky struktury FILE skryté spočívá v tom, že
      // soubory mohou být v různých operačních systémech realizovány různým
      // způsobem. Například Linux používá jiný systém práv přístupu než Windows.
      // Liší se také vnitřní datová struktura pro ukládání souborů - Linux
      // v systému ext3, ext4 používá tzv. inody, které tvoří stromovou
      // strukturu, systém FAT je tvořen lineární tabulkou a existuje množství
      // dalších rozdílů.
      // Na různých systémech bude struktura FILE vypadat jinak.
      // Pokud by měl uživatel do struktury přístup, mohl by snadno něco poškodit.
      "  - Vytváří abstrakci čtecí/zapisovací hlavy - pamatuje si, kde skončila\n"
      "    práci poslední operace, aby odsud mohla další operace pokračovat.\n"
      "\n"
      "- Všechny potřebné operace se souborem dělají funkce z modulu " KK("stdio.h") ".\n"
      "- Všechny tyto funkce pracují s typem " KK("FILE *") ".\n"
      "\n"
    );
  }

  if (navratNeboSekce("Textový a binární soubor")) return;
  if (pokracovat())
  {
    printh(
      "- Neexistuje principiální rozdíl mezi textovým a binárním souborem.\n"
      "- Vždy jde o sekvenční posloupnost bajtů.\n"
      "\n"
      "- Ve Windows se binární soubory otevírají jinak než textové kvůli\n"
      "  problému s konci řádků.\n"
      "- Ve Windows konec řádku tvoří dva znaky:\n"
      "  CR - Carriage return - návrat vozíku - \\r (ASCII kód: %d) a\n"
      "  LF - Line feed - posun papíru o řádek výše - \\n (ASCII kód: %d).\n"
      "- Tyto znaky byly původně příkazy pro dálnopis z 1. půlky 20. století.\n"
      "\n"
      "- Otevře-li se ve Windows soubor jako textový, funkce pro čtení tyto dva\n"
      "  znaky přečtou a vrátí jako jeden znak \\n. Při zápisu se pak znak \\n\n"
      "  zapíše jako dva znaky \\n\\r.\n"
      "- Otevře-li se ve Windows soubor jako binární, tyto dva znaky se čtou\n"
      "  i zapisují samostatně a nepřekládají se.\n"
      "\n"
      "- V unixu je konec řádku reprezentován jen jedním znakem \\n (ASCII: %d).\n"
      "  => V unixu není potřeba otvírat textový a binární soubor jinak.\n"
      "\n"
      , '\r', '\n'
      , '\n'
    );
  }

  if (navratNeboSekce("Výhody textových souborů")) return;
  if (pokracovat())
  {
    printh(
      "- Textový formát řeší mnoho problémů s kompatibilitou.\n"
      "- Text je přímo čitelný a jednoduše editovatelný.\n"
      "- CSV formát (Comma-separated Values) lze snadno zpracovávat například\n"
      "  v tabulkových procesorech.\n"
      "- CSV formát lze snadno zpracovávat i standardními funkcemi jazyka C\n"
      "   pro formátovaný zápis a čtení.\n"
      "- Textový formát je snadno komprimovatelný.\n"
      "- Nehodí se však na ukládání velkých objemů surových dat jako jsou\n"
      "  multimediální data - zvuk, video, fotografie.\n"
      "\n"
      "- Binární soubory vyžadují mnohem větší programátorské zkušenosti.\n"
      "- Aby byly přenositelné, musí řešit i hardwarové problémy, například\n"
      "  problém little a big endian, tedy pořadí uložení bajtů vícebajtových\n"
      "  číselných hodnot v paměti.\n"
      "\n"
    );
  }

  if (navratNeboSekce("Konec souboru")) return;
  if (pokracovat())
  {
    printh(
      "- Na dosažení konce souboru čtecí funkce reagují tak, že vrátí funkční\n"
      "  hodnotu " KK("EOF") " (nebo " KK("WEOF") " při širokém, Unicode módu).\n"
      "- " KK("EOF") " je konstanta typu " KK("int") ".\n"
      "- V jazyce C nejde předem zjistit, že při dalším čtení následuje konec souboru.\n"
      // Tohle umí jazyk Pascal a Pascalisti při přechodu na C s tím mívali problém...
      "- V jazyce C lze na konci souboru provést operaci čtení a nezpůsobí to chybu.\n"
      "  => Funkce prostě zase vrátí hodnotu " KK("EOF") ".\n"
      "- Stav dosažení konce souboru testuje i funkce " KK("feof") ".\n"
      "\n"
      "- Pozor! Na konci souboru NENÍ zapsán žádný speciální znak nebo značka\n"
      "  s hodnotou " KK("EOF") "!\n"
      "- Informaci o dosažení konce souboru dává souborový systém - ví, jak je\n"
      "  soubor dlouhý.\n"
      "- Soubory se značkou na konci nemohou z principu fungovat!\n"
      "  => Hlavně u binárních souborů by se taková značka mohla objevit náhodně\n"
      "     v uložených datech a nezamýšleně by to soubor ukončilo.\n"
      // Data jako fotky nebo video jsou v podstatě náhodné bajty, ve kterých
      // se s nenulovou pravděpodobností může vyskytnout jakákoli posloupnost,
      // tedy i taková, která by měla význam koncové značky.
      "\n"
      "- Kvůli tomu, že " KK("EOF") " je typu " KK("int") ", funkce jako " KK("getchar") " vrací také hodnoty\n"
      "  typu " KK("int") ".\n"
      "- Kdyby vracela " KK("char") ", jeden ze znaků by byl nerozeznatelný od " KK("EOF") ".\n"
      "  => Při čtení po znacích se musí pracovat s proměnnou typu " KK("int") ", ne " KK("char") ".\n"
      "\n"
    );
  }

  if (navratNeboSekce("Sekvenční a náhodný přístup")) return;
  if (pokracovat())
  {
    printh(
      "- Soubor považujeme za sekvenční datovou strukturu.\n"
      "\n"
      PODSEKCE("Sekvenční přístup")
      "- Soubor se čte po bajtech postupně od začátku do konce.\n"
      "- Je nutný kvůli\n"
      "  - sekvenčním paměťovým médiím, např. magnetickým páskám,\n"
      "  - datovým proudům (stream),\n"
      "  - fragmentaci souborů - náhodný přístup by byl příliš složitý.\n"
      "=> Standardní souborové rozhraní realizuje sekvenční přístup.\n"
      "\n"
      PODSEKCE("Náhodný přístup")
      "- Znamená to možnost okamžitého přístupu k libovolnému bajtu souboru.\n"
      "- Teoreticky to umí moderní paměťová média jako SSD.\n"
      "- Je potřeba použít speciální funkce, jako " KK("fseek") ", které ale\n"
      "  na pomalých médií budou extrémně pomalé nebo nemusí fungovat vůbec.\n"
      "- Některé datové proudy tento přístup z principu nemohou realizovat.\n"
      // Například vstup z klávesnice nemůže předpovídat budoucnost a přečíst
      // znak, který ještě nebyl zapsán.
      "\n"
    );
  }

  konecKapitoly(__func__);
}



////////////////////////////////////////////////////////////////////////////////
void datovyProud(void)
{
  zacatekKapitoly("Datový proud - stream");

  printh(
    PODSEKCE("Soubor")
    "- Tvoří ho data uložená na paměťovém médiu.\n"
    "- Má konečnou velikost.\n"
    "- Může podporovat náhodný přístup.\n"
    "\n"
    PODSEKCE("Datový proud")
    "- Je to abstrakce tvořená operačním systémem (OS).\n"
    "- Netvoří jej konkrétní data uložená na paměťovém médiu.\n"
//      "- OS pro ně ale může vytvořit tzv. virtuální soubory v souborovém systému.\n"
//      "  => Např. v Linuxu existuje \"soubor\" /dev/stdout.\n"
    "- Nemá předem zjistitelnou délku - může být \"nekonečný\".\n"
    "\n"
    PODSEKCE("Datové proudy pro čtení")
    "- OS na ně mapuje některá zařízení, která kontinuálně posílají data.\n"
    "- Myš, klávesnice, kamera, mikrofon, různé senzory, síťové rozhraní aj.\n"
    "- Z principu musí fungovat sekvenčně.\n"
    "=> Nesekvenční přístup ke klávesnici by musel umět číst budoucnost...\n"
    "\n"
    PODSEKCE("Datové proudy pro zápis.")
    "- Obrazovka, textová konzole, reproduktory, síťové rozhraní aj.\n"
    // Tento přístup je typický zejména pro unixové systémy (Linux, MacOS).
    // Ve Windows funguje taky, ale je více ukrytý před uživateli.
    "\n"
    PODSEKCE("Datové proudy pro čtení a zápis")
    "- Říká se jim duplexní proudy.\n"
//    "- Mohou to být například síťové sokety nebo pojmenované roury pro komunikaci\n"
//    "  mezi více procesy.\n"
//    "- Jsou to však systémově specifické a nekompatibilní řešení. Standard\n"
//    "  jazyka C duplexní proudy nezahrnuje.\n"
    "\n"
  );

  if (navratNeboSekce("Standardní datové proudy")) return;
  if (pokracovat())
  {
    printh(
      "- Knihovna jazyka C definuje tři globální proměnné typu " KK("FILE *") ",\n"
      "  které reprezentují standardní datové proudy pro běžné použití.\n"
      "- Tyto proudy se tváří jako soubory otevřené pro čtení, nebo zápis.\n"
      "- Nesmí se otvírat/zavírat pomocí " KK("fopen") "/" KK("fclose") ".\n"
      "\n"
      PODSEKCE("stdin")
      "- Standardní vstup -> Čte z něj " KK("getchar") ", " KK("scanf") ".\n"
      "- Data z klávesnice\n"
      "- Data přesměrovaná ze souboru nebo roury při spuštění programu\n"
      "  => " KK("program.exe < vstup.txt") "\n"
      "\n"
      PODSEKCE("stdout")
      "- Standardní výstup -> Zapisuje sem " KK("putchar") ", " KK("printf") ".\n"
      "- Data běžně tisknutá na textovou konzoli\n"
      "- Jde jej přesměrovat do souboru nebo roury při spuštění programu.\n"
      "  => " KK("program.exe > vystup.txt") "\n"
      "  => " KK("program.exe | jiny-program.exe") "\n"
      "\n"
      PODSEKCE("stderr")
      "- Standardní chybový výstup -> Pro tisk chybových hlášení\n"
      "- Běžně se objevuje na textové konzoli.\n"
      "- Jde jej přesměrovat do souboru při spuštění programu a oddělit tak\n"
      "  běžný výstup od chybových zpráv.\n"
      "  => " KK("program.exe > vystup.txt 2> chybova-hlaseni.txt") "\n"
      "\n"
      "- Standardní funkce pro čtení/tisk z/do konzole pracují s těmito proudy.\n"
      "- Funkce pro čtení/tisk z/do souboru vypadají a fungují záměrně stejně.\n"
      // Vnitřně je například funkce printf realizována voláním fprintf(stdout, ...).
      "\n"
    );
  }

  if (navratNeboSekce("Ukázka - zápis do stdout")) return;
  if (pokracovat())
  {
    printh(
      "- Funkce " KK("printf") " pro tisk na konzoli ve skutečnosti zapisuje do " KK("stdout") ".\n"
      "- Funkce " KK("fprintf") " zapisuje do souboru nebo proudu. Dostane-li jako\n"
      "  argument " KK("stdout") " místo souboru, tiskne na konzoli.\n"
      "\n"
      KOD(
      "  printf(\"  => Tisknu na konzoli pomoci printf.\\n\");\n"
      "  fprintf(stdout,\"  => Tisknu do stdout pomoci fprintf.\\n\");\n"
      )
      "\n"
    );

    printf("  => Tisknu na konzoli pomoci printf.\n");
    fprintf(stdout,"  => Tisknu do stdout pomoci fprintf.\n");

    printh(
      "\n"
      "- Tip: Pokud potřebuješ tisknout data stejně do souboru a na konzoli\n"
      "  není potřeba vyrábět dvě funkce. Vyrob funkci pro tisk do souboru\n"
      "  a tisk na obrazovku realizuj voláním s argumentem " KK("stdout") ".\n"
      "\n"
    );
  }

  if (navratNeboSekce("Ukázka - čtení ze stdin")) return;
  if (pokracovat())
  {
    printh(
      "- Funkce " KK("scanf") " ve skutečnosti čte ze " KK("stdin") ".\n"
      "- Stejně jde použít i funkce " KK("fscanf") ".\n"
      "- Při zadávání dat z klávesnice se proud ukončuje klávesovou zkratkou:\n"
      "  - Ctrl+D v unixových systémech (Linux, MacOS)\n"
      "  - Ctrl+Z ve Windows\n"
      // Pozor! V Linuxu zkratka Ctrl+Z posílá aktivní program na pozadí.
      // To pro nezkušeného uživatele může vypadat jako vypnutí programu, který
      // ale pořád běží. Na popředí jej jde v konzoli vrátit příkazem fg.
      "\n"
      KOD(
      "  fprintf(stdout, \"  Zadej text (konec - Ctrl+D v Linuxu, Ctrl+Z ve Windows):\\n\");\n"
      "  char slovo[50+1];\n"
      "  int kod;\n"
      "  while ( (kod = fscanf(stdin, \"%%50s\", slovo))  == 1 )\n"
      "  {\n"
      "    printf(\"  => slovo: |%%s|, kod: %%d\\n\", slovo, kod);\n"
      "  }\n"
      "\n"
      "  clearerr(stdin); // Smaže příznak dosažení konce souboru/streamu.\n"
      )
      "\n"
    );

    char znovu = 'z';
    while (znovu == 'z')
    {
      fprintf(stdout, "  Zadej text (konec - Ctrl+D v Linuxu, Ctrl+Z ve Windows):\n");
      char slovo[50+1];
      int kod;
      while ( (kod = fscanf(stdin, "%50s", slovo))  == 1 )
      {
        printf("  => slovo: |%s|, kod: %d\n", slovo, kod);
      }

      clearerr(stdin); // Smaže příznak dosažení konce souboru/streamu.
      // Bez toho by už program nepřijímal žádný vstup.

      printg(VOLBA("## Znovu? -> Z, Pokračuj -> Enter: "));
      znovu = tolower(prvniPismenoNaRadku());
    }

  }

  konecKapitoly(__func__);
}





////////////////////////////////////////////////////////////////////////////////

/** Položí otázku na cestu k souboru a přečte odpověď.
 *
 * Umí načíst i cestu s mezerami.
 *
 * \param otazka const char* Text otázky pro uživatele
 * \param cesta[] char Pole znaků, kam se zadaná cesta uloží.
 * \param delka int Počet znaků, které se maximálně načtou do pole cesta.
 * \return void
 */
int otazkaStr(const char * otazka, char cesta[], int delka)
{
  printf("%s", otazka);
  char format[20+1];
  sprintf(format, " %%%d[^\n]", delka);
  return scanf(format, cesta);
}

/** Přeskočí bílé znaky a testuje, jestli narazila na konec řádku.
 *
 * \param f FILE* Ukazatel na otevřený soubor
 * \return bool Vrací true, když narazila na EOF nebo konec řádku, jinak false.
 *
 */
bool konecRadku(FILE *f)
{
  int znak;
  while ( isblank(znak = fgetc(f)) ) // přeskočí mezery a tabulátory
  {}

  if (znak == '\n' || znak == EOF)
    return true;

  ungetc(znak, f); // Protože se načetl první znak slova, vrátím ho zpět.
  return false;
}

/** Přečte slova z jednoho řádku vstupního souboru.
 *
 * \param f FILE* Ukazatel na otevřený soubor.
 * \return int Vrací kód, posledního čtení pomocí fscanf => na konci vrací EOF.
 */
int zpracujRadek(FILE *f)
{
  char slovo[20+1];
  int kod = 1;

  while (!konecRadku(f) && kod != EOF)
  {
    kod = fscanf(f, "%20s", slovo);
    printf("  => slovo: |%s|, kod: %d\n", slovo, kod);
  }
  return kod; // Vrací případný EOF, kdyby bylo potřeba to volat v cyklu.
}

void pripravaSouboru(void)
{
  zacatekKapitoly("Příprava souboru pro práci");

  printh(
    "- Soubor je potřeba před zpracováním otevřít v zamýšleném módu.\n"
    "- Na konci je nutné soubor správně zavřít.\n"
    "- Je potřeba ošetřovat případné chyby, jinak dochází k haváriím\n"
    "  a co hůře - k poškození zpracovávaného souboru!\n"
    "\n"
  );

  if (navratNeboSekce("Schéma práce se souborem")) return;
  if (pokracovat())
  {
    printh(
      "- Práce se souborem MUSÍ obsahovat všechny tyto 4 kroky:\n"
      "  1. " KK("fopen") " - Pokus o otevření souboru\n"
      "  2. Test, pokud fopen vrací " KK("NULL") " => soubor se nepovedlo otevřít.\n"
      "  3. Práce se souborem\n"
      "  4. " KK("fclose") " - Zavření souboru\n"
      "- Ideální je, když se " KK("fopen") " a " KK("fclose") " volají ve stejném bloku,\n"
      "  například ve funkci main.\n"
      "\n"
      KOD(
      "  // 1. fopen - Pokus o otevření souboru\n"
      "  FILE * soubor = fopen(cesta, \"r\"); // Otevírám pro čtení.\n"
      "\n"
      "  // 2. Test NULL. - Co když se otevření nepovedlo?\n"
      "  if (soubor == NULL)\n"
      "  {\n"
      "    perror(\"Chyba pri otevirani souboru\"); // Tisk chybového hlášení na stderr.\n"
      "    return; // Končím. Nic nezavírám - soubor není otevřený.\n"
      "  }\n"
      "\n"
      "  // 3. Práce se souborem\n"
      "  zpracujRadek(soubor);\n"
      "\n"
      "  // 4. fclose - Zavření souboru\n"
      "  fclose(soubor);\n"
      )
      "\n"
    );
    char znovu = 'z';
    while (znovu == 'z')
    {
      char cesta[255+1];
      if (otazkaStr("  Zadej cestu k souboru: (zkus data/data00.txt):\n  ", cesta, 255) != 1)
      {
        printf("Neco se pokazilo. Zkusim cestu data/data00.txt.\n");
        strcpy(cesta, "data/data00.txt");
      }

      // 1. fopen - pokus o otevření souboru
      FILE * soubor = fopen(cesta, "r");
      // 2. testuj NULL - co když se otevření nepovedlo?
      if (soubor == NULL)
      {
        perror("Chyba pri otevirani souboru"); // Tisk chybového hlášení na stderr.
        //return; // Končím. Nic nezavírám - soubor není otevřený.

        goto konecCyklu; // Fuj! Tohle je špinavý hack!
        // Lepší by bylo celou ukázku zabalit do funkce, kde by to končilo
        // příkazem return. Ale aby ta ukázka byla názorně tady, tak si hold
        // pomůžu tímhle prehistorickým hackem...
        // Vidíte, že když to pořádně zdůvodníte a obhájíte, jde výjimečně použít
        // i takovouto prasárnu. ;-)
      }
      // 3. zpracuj soubor
      zpracujRadek(soubor);
      // 4. fclose - zavři soubor
      fclose(soubor);

konecCyklu:
      pozerKonecRadku(); // Po zadání cesty asi zůstal na vstupu znak \n.
      printg(VOLBA("## Znovu? -> Z, Pokračuj -> Enter: "));
      znovu = tolower(prvniPismenoNaRadku());
    }
  }

  if (navratNeboSekce("Funkce fopen")) return;
  if (pokracovat())
  {
    printh(
      "- Funkce " KK("fopen") " otevírá soubor se zadanou cestou v zadaném módu.\n"
      "- Vrací ukazatel na popisovač souboru nebo " KK("NULL") ".\n"
      "- Návratovou hodnotu je NUTNÉ testovat.\n"
      "  => Práce s neotevřeným souborem způsobí havárii programu!\n"
      "- Vnitřně funkce volá " KK("malloc") ", funkce " KK("fclose") " pak " KK("free") ".\n"
      "\n"
      KOD(
      "  FILE *fopen(const char *cesta, const char *mod);\n"
      )
      "\n"
      PODSEKCE("Cesta")
      "- Cesta k otevíranému souboru.\n"
      "- Pokud je soubor ve stejné složce, jako program, stačí jen jméno souboru.\n"
      "- Zvládá jak relativní, tak absolutní cesty.\n"
      "- V unixových systémech (Linux, MacOS) je oddělovačem adresářů lomítko (/).\n"
      "- Ve Windows je oddělovačem adresářů opačné lomítko (\\).\n"
      // MS musí dělat všechno po svém...
      "- Moderní OS dovolují v cestě i názvu souboru mezery a lokalizované znaky.\n"
      "  Pokud je dovedeš od uživatele přečíst, je to v pořádku.\n"
      "  Pokud ne, raději se mezerám a českým znakům v názvech složek i souborů\n"
      "  vyhýbej!\n"
      "\n"
    );
  }

  if (navratNeboSekce("Základní módy práce se souborem")) return;
  if (pokracovat())
  {
    printh(
      "- Pozor! Parametr " KK("mod") " je string, i když jej často tvoří jeden znak.\n"
      "  => Vždy používej dvojité uvozovky, ne jednoduché!\n"
      "\n"
      "" KK("\"r\"") "\n"
      "- Nedestruktivní otevírání pro čtení\n"
      "- Vrací " KK("NULL") ",\n"
      "  - když soubor neexistuje,\n"
      "  - když program nemá právo číst ze souboru.\n"
      "\n"
      "" KK("\"w\"") "\n"
      "- Destruktivní otevírání pro zápis\n"
      "- Pokud soubor existuje, smaže se a vytvoří se nový soubor.\n"
      "- Pokud soubor neexistuje, vytvoří se nový.\n"
      "- Vrací " KK("NULL") ",\n"
      "  - když program nemá právo vytvářet soubor,\n"
      "  - když program nemá právo zapisovat do souboru.\n"
      "\n"
      "" KK("\"a\"") "\n"
      "- Nedestruktivní otevírání pro zápis na konec souboru\n"
      "- Pokud soubor neexistuje, vytvoří se nový.\n"
      "- Vrací " KK("NULL") ",\n"
      "  - když program nemá právo vytvářet soubor,\n"
      "  - když program nemá právo zapisovat do souboru.\n"
      "\n"
    );
  }

  if (navratNeboSekce("Rozšířené módy práce se souborem")) return;
  if (pokracovat())
  {
    printh(
      "- Umožňují zároveň čtení i zápis.\n"
      "- Předpokládá se použití " KK("fseek") ".\n"
      "\n"
      "" KK("\"r+\"") "\n"
      "- Jako mód " KK("\"r\"") " + umožní zápis.\n"
      "- Nedestruktivní otevírání pro čtení a zápis do souboru.\n"
      "- Čtecí hlava je nastavena na začátek souboru.\n"
      "\n"
      "" KK("\"w+\"") "\n"
      "- Jako mód " KK("\"w\"") " + umožní čtení.\n"
      "- Destruktivní otevírání pro zápis a čtení ze souboru.\n"
      "- Čtecí hlava je nastavena na začátek souboru.\n"
      "\n"
      "" KK("\"a+\"") "\n"
      "- Jako mód " KK("\"a\"") " + umožní čtení.\n"
      "- Nedestruktivní otevírání pro zápis na konec souboru a čtení.\n"
      "- Čtecí hlava je nastavena na konec souboru.\n"
      "\n"
      "" KK("\"b\"") "\n"
      "- Otvírá soubor jako binární.\n"
      "- Má smysl jen ve Windows - viz rozdíl mezi textovým a binárním souborem.\n"
      "- Unixové systémy jej ignorují.\n"
      "- Používá se jako přípona k předchozím módům: " KK("rb") ", " KK("w+b") ", " KK("wb+") ", ...\n"
      "\n"
    );
  }

  if (navratNeboSekce("Funkce fclose")) return;
  if (pokracovat())
  {
    printh(
      "\n"
      KOD(
      "  int fclose(FILE *soubor);\n"
      )
      "\n"
      "- Funkce " KK("fclose") " zavře otevřený soubor.\n"
      "- Před zavřením vyprázdní vnitřní řádkový buffer.\n"
      "  => Pokud se soubor v módu pro zápis nezavře, může dojít k nezapsání\n"
      "     části dat, tj. poškození souboru.\n"
      "\n"
      "- Funkce " KK("fclose") " vnitřně používá " KK("free") " pro uvolnění dynamicky alokované\n"
      "  datové struktury.\n"
      "  => Program, který nezavírá soubory, má paměťové úniky (memory leaks).\n"
      "\n"
      "- Zavřít jde pouze úspěšně otevřený soubor. Pokus o zavření neotevřeného\n"
      "  nebo již zavřeného souboru způsobí havárii programu.\n"
      "\n"
    );
  }

  konecKapitoly(__func__);
}



////////////////////////////////////////////////////////////////////////////////
void testovaniAZpracovaniChyb(void)
{
  zacatekKapitoly("Testování a zpracování chyb");

  printh(
    "- Při práci se soubory je důležité testovat správnost provedení všech\n"
    "  operací.\n"
    "- Program, který je netestuje => havaruje.\n"
    "\n"
    "- Všechny I/O operace nějakým způsobem vracejí informaci, zda proběhly\n"
    "  správně, nebo došlo k chybě.\n"
    "  - " KK("fscanf") " vrací počet úspěšně zpracovaných značek\n"
    "  - " KK("fopen") " vrací v případě neúspěchu " KK("NULL") "\n"
    "  - " KK("fgetc") " vrací při chybě nebo na konci souboru " KK("EOF") "\n"
    "  - Další viz manuálové stránky.\n"
    "- Většinou to pro určení zdroje problému stačí.\n"
    "\n"
    "- Pro přesnější určení chyby slouží proměnná " KK("errno") " a funkce \n"
    "  " KK("perror") ", " KK("strerror") ".\n"
    "\n"
  );

  if (navratNeboSekce("Vnitřní kód chyby errno")) return;
  if (pokracovat())
  {
    printh(
      "- V modulu errno.h je vytvořena globální proměnná " KK("errno") " typu " KK("int") ".\n"
      "- Všechny I/O operace do ní zapisují kód chyby.\n"
      "  => Tato proměnná se musí zjišťovat bezprostředně po provedené operaci.\n"
      "  => Jiná I/O operace ji může přenastavit.\n"
      "\n"
      KOD(
      "  FILE * f = fopen(\"blabol.ble\", \"r\");\n"
      "  int kodChyby = errno;\n"
      "  printf(\"  => f: %%p, kodChyby: %%d\\n\", (void*)f, kodChyby);\n"
      )
      "\n"
    );

    FILE * f = fopen("blabol.ble", "r");
    int kodChyby = errno;
    printf("  => f: %p, kodChyby: %d\n\n", (void*)f, kodChyby);

  }

  if (navratNeboSekce("Funkce perror")) return;
  if (pokracovat())
  {
    printh(
      "- Tato funkce vytiskne na chybový výstup chybové hlášení odpovídající\n"
      "  aktuální hodnotě proměnné " KK("errno") ".\n"
      "  => Musí se opět volat bezprostředně po chybné operaci.\n"
      "- Jako argument se zadává vlastní hlášení.\n"
      "\n"
      "- Pokud je zapnutá česká lokalizace pomocí " KK("setlocale") ", vypisuje\n"
      "  hlášení česky, jinak anglicky.\n"
      "\n"
      KOD(
      "  FILE * f = fopen(\"blabol.ble\", \"r\");\n"
      "  perror(\"  => Vysledek operace fopen\");\n"
      )
      "\n"
    );

    FILE * f = fopen("blabol.ble", "r");
    perror("  => Vysledek operace fopen");
    if (f != NULL) fclose(f);
    printf("\n");

  }

  if (navratNeboSekce("Funkce strerror")) return;
  if (pokracovat())
  {
    printh(
      "- Vrací ukazatel na string s chybovým hlášením odpovídajícím chybovému kódu.\n"
      "- Sama proměnnou " KK("errno") " nepoužívá. Tuto hodnotu je třeba předat jako\n"
      "  argument.\n"
      "\n"
      "- Vrací ukazatel na konstantní string. => Nesmí se modifikovat.\n"
      "- Funkce je v modulu string.h.\n"
      "\n"
      KOD(
      "  FILE * f = fopen(\"blabol.ble\", \"r\");\n"
      "  int kodChyby = errno;\n"
      "  printf(\"  => f: %%p, kodChyby: %%d\\n\", (void*)f, kodChyby);\n"
      "  printf(\"  => chybove hlaseni: %%s\\n\", strerror(kodChyby));\n"
      )
      "\n"
    );

    FILE * f = fopen("blabol.ble", "r");
    int kodChyby = errno;
    printf("  => f: %p, kodChyby: %d\n", (void*)f, kodChyby);
    printf("  => chybove hlaseni: %s\n\n", strerror(kodChyby));
  }

  if (navratNeboSekce("Funkce feof, ferror")) return;
  if (pokracovat())
  {
    printh(
      "\n"
      KOD(
      "  int feof(FILE *f);\n"
      "  int ferror(FILE *f);\n"
      )
      "\n"
      "- Tyto funkce vrací logickou hodnotu 0 nebo 1.\n"
      "- Funkce " KK("feoff") " testuje, zda má zadaný soubor čtecí hlavu za koncem souboru.\n"
      "  - Pozor! Nejde zjistit, že bajt, který jsem teď přečetl je poslední.\n"
      "  - Soubor se do tohoto stavu dostane až následujícím pokusem o čtení.\n"
      "\n"
      "- Funkce " KK("ferror") " testuje, zda je soubor v chybovém stavu.\n"
      "  - Pozor! Nejde použít pro test po " KK("fopen") ", protože by argument byl " KK("NULL") ".\n"
      "\n"
      "- Poznámka: Tyto funkce obvykle není potřeba používat. Předchozí zmíněné\n"
      "  způsoby jsou dostačující.\n"
      "\n"
    );

  }

  if (navratNeboSekce("Funkce clearerr")) return;
  if (pokracovat())
  {
    printh(
      "- Smaže zadanému proudu nebo souboru příznaky dosažení konce souboru\n"
      "  a příznak chyby.\n"
      "- Používá se pro zotavení po některých chybách.\n"
      "- Dá se například použít pro znovuoživení proudu stdin, když jej uživatel\n"
      "  ručně ukončil klávesovou zkratkou (Ctrl-D v Linuxu, Ctrl-Z ve Windows).\n"
      "\n"
      KOD(
      "  fprintf(stdout, \"  Zadej text (konec - Ctrl+D v Linuxu, Ctrl+Z ve Windows):\\n\");\n"
      "  char slovo[50+1];\n"
      "  int kod;\n"
      "  while ( (kod = fscanf(stdin, \"%%50s\", slovo))  == 1 )\n"
      "  {\n"
      "    printf(\"  => slovo: |%%s|, kod: %%d\\n\", slovo, kod);\n"
      "  }\n"
      "\n"
      "  clearerr(stdin); // Smaže příznak dosažení konce souboru/streamu.\n"
      )
      "\n"
    );

    char znovu = 'z';
    while (znovu == 'z')
    {
      fprintf(stdout, "  Zadej text (konec - Ctrl+D v Linuxu, Ctrl+Z ve Windows):\n");
      char slovo[50+1];
      int kod;
      while ( (kod = fscanf(stdin, "%50s", slovo))  == 1 )
      {
        printf("  => slovo: |%s|, kod: %d\n", slovo, kod);
      }

      clearerr(stdin); // Smaže příznak dosažení konce souboru/streamu.
      // Bez toho by už program nepřijímal žádný vstup.

      printg(VOLBA("## Znovu? -> Z, Pokračuj -> Enter: "));
      znovu = tolower(prvniPismenoNaRadku());
    }
  }

  konecKapitoly(__func__);
}



////////////////////////////////////////////////////////////////////////////////
void zapisDoSouboru(void)
{
  zacatekKapitoly("Zápis do textových souborů a proudů");

  printh(
    "- Soubor musí být otevřený v některém z módů w, a, r+, w+, a+.\n"
    "- Zápis jde také provádět do datových proudů pro zápis, jako " KK("stdout") " a " KK("stderr") ".\n"
    "\n"
    "- Rozhraní pro zápis do souboru je téměř úplně totožné s rozhraním pro\n"
    "  zápis na standardní výstup (" KK("stdout") "). Je to záměr.\n"
    "- Jeden rozdíl je v názvu funkcí - typicky mají v názvu předponu f.\n"
    "  => " KK("printf") " -> " KK("fprintf") ".\n"
    "- Druhým rozdílem je parametr typu " KK("FILE *") ".\n"
    "\n"
  );

  if (navratNeboSekce("Zápis po znacích - fputc")) return;
  if (pokracovat())
  {
    printh(
      "- Funkce " KK("fputc") " se chová stejně jako " KK("putchar") ", jen tiskne do souboru.\n"
      "- Tyto dva příkazy dělají totéž:\n"
      "\n"
      KOD(
      "  putchar('A');\n"
      "  fputc('A', stdout);\n"
      )
      "\n"
      "- Tiskne znak uložený do typu " KK("int") ".\n"
      "- Vrací úspěšně zapsaný znak, nebo hodnotu " KK("EOF") " při chybě.\n"
      "\n"
    );
  }

  if (navratNeboSekce("Ukázka - fputc")) return;
  if (pokracovat())
  {
    printh(
      "\n"
      KOD(
      "  char * cesta = \"data/data01.txt\";\n"
      "  FILE * f = fopen(cesta, \"w\");\n"
      "  if (f != NULL)\n"
      "  {\n"
      "    char *text = \"Sla Nanynka do zeli, natrhala lupeni.\\n\";\n"
      "    for (int i = 0; text[i] != 0; ++i)\n"
      "    { fputc(text[i], f); }\n"
      "    fclose(f);\n"
      "\n"
      "    printg(\"  => Podívej se do souboru %%s.\\n\", cesta);\n"
      "  }\n"
      "  else\n"
      "  { perror(\"Problem se souborem data01.txt\"); }\n"
      )
      "\n"
    );

    char * cesta = "data/data01.txt";
    FILE * f = fopen(cesta, "w");
    if (f != NULL)
    {
      char *text = "Sla Nanynka do zeli, natrhala lupeni.\n";
      for (int i = 0; text[i] != 0; ++i)
      { fputc(text[i], f); }
      fclose(f);

      printg("  => Podívej se do souboru %s.\n", cesta);
    }
    else
    { perror("Problem se souborem data01.txt"); }

    printf("\n");
  }

  if (navratNeboSekce("Formátovaný zápis - fprintf")) return;
  if (pokracovat())
  {
    printh(
      "- Funkce " KK("fprintf") " se chová stejně jako " KK("printf") ", jen tiskne do souboru.\n"
      "\n"
      KOD(
      "...\n"
      "  fprintf(f, \"%%20s;%%10d;%%20.3f\\n\", \"Alfons\", 2148, 638.9742);\n"
      "  fprintf(f, \"%%20s;%%10d;%%20.3f\\n\", \"Aurelie\", 9762, 2486.6571);\n"
      "  fprintf(f, \"%%20s;%%10d;%%20.3f\\n\", \"Izidor\", 7564, 12.36);\n"
      "...\n"
      )
      "\n"
    );

    char * cesta = "data/data02.txt";
    FILE * f = fopen(cesta, "w");
    if (f != NULL)
    {
      fprintf(f, "%20s;%10d;%20.3f\n", "Alfons", 2148, 638.9742);
      fprintf(f, "%20s;%10d;%20.3f\n", "Aurelie", 9762, 2486.6571);
      fprintf(f, "%20s;%10d;%20.3f\n", "Izidor", 7564, 12.36);
      fclose(f);
      printg("  => Podívej se do souboru %s.\n", cesta);
    }
    else
    { perror("Problem se souborem data02.txt"); }

    printf("\n");
  }

  if (navratNeboSekce("Zápis do širokých souborů")) return;
  if (pokracovat())
  {
    printh(
      "- Jde pracovat i se soubory v Unicode kódování.\n"
      "- Každý soubor je potřeba přepnout do širokého módu.\n"
      "  Viz kapitolu Textové řetězce.\n"
      "- Široký mód se u každého souboru či proudu zapíná individuálně.\n"
      "  => Není problém pracovat se širokými soubory, ale na konzoli pracovat\n"
      "     v bajtovém módu.\n"
      "- Ke všem bajtovým funkcím existují široké alternativy, například\n"
      "  " KK("fputwc") ", " KK("fwprintf") ", " KK("fputws") ".\n"
      "\n"
    );
  }

  if (navratNeboSekce("Ukázka - použití fwprintf")) return;
  if (pokracovat())
  {
    printh(
      "- Pozor! Funkce " KK("fwprintf") " rozlišuje mezi značkami " KK("%%s") " a " KK("%%ls") "\n"
      "  pro obyčejné a široké stringy.\n"
      "\n"
      KOD(
      "...\n"
      "  if (fwide(f, +1) > 0) // Na začátku přepnu f do širokého režimu.\n"
      "  {\n"
      "    fwprintf(f, L\"%%20ls;%%10d;%%20.3f\\n\", L\"Příliš\", 2148, 638.9742);\n"
      "    fwprintf(f, L\"%%20ls;%%10d;%%20.3f\\n\", L\"Žluťoučký\", 9762, 2486.6571);\n"
      "    fwprintf(f, L\"%%20ls;%%10d;%%20.3f\\n\", L\"Kůň\", 7564, 12.36);\n"
      "  }\n"
      "...\n"
      )
      "\n"
    );

    char * cesta = "data/wdata03.txt";
    FILE * f = fopen(cesta, "w");
    if (f != NULL)
    {
      if (fwide(f, +1) > 0) // Na začátku přepnu f do širokého režimu.
      // Není nutné soubor přepínat takto explicitně, udělá to i první použití
      // fwprintf. U funkce fwide ale jde otestovat, jestli se to povedlo.
      {
        fwprintf(f, L"%20ls;%10d;%20.3f\n", L"Příliš", 2148, 638.9742);
        fwprintf(f, L"%20ls;%10d;%20.3f\n", L"Žluťoučký", 9762, 2486.6571);
        fwprintf(f, L"%20ls;%10d;%20.3f\n", L"Kůň", 7564, 12.36);
      }
      fclose(f);
      printg("  => Podívej se do souboru %s.\n", cesta);
    }
    else
    { perror("Problem se souborem wdata03.txt"); }

    printf("\n");
  }

  konecKapitoly(__func__);
}



////////////////////////////////////////////////////////////////////////////////
void cteniZeSouboru(void)
{
  zacatekKapitoly("Čtení z textových souborů a proudů");

  printh(
    "- Soubor musí být otevřený v některém z módů r, r+, w+, a+.\n"
    "- Číst jde také z datových proudů pro čtení, jako " KK("stdin") ".\n"
    "\n"
    "- Rozhraní pro čtení ze souboru je téměř úplně totožné s rozhraním pro\n"
    "  čtení ze standardního vstupu (" KK("stdin") "). Je to záměr.\n"
    "- Jeden rozdíl je v názvu funkcí - typicky mají v názvu předponu f.\n"
    "  => " KK("scanf") " -> " KK("fscanf") ".\n"
    "- Druhým rozdílem je parametr typu " KK("FILE *") ".\n"
    "\n"
  );

  if (navratNeboSekce("Čtení po znacích")) return;
  if (pokracovat())
  {
    printh(
      "- Funkce " KK("fgetc") " se chová stejně jako " KK("getchar") ", jen čte ze souboru.\n"
      "- Tyto dva příkazy dělají totéž:\n"
      "\n"
      KOD(
      "  int znak = getchar();\n"
      "  znak = fgetc(stdin);\n"
      )
      "\n"
      "- Čte znak a vrací ho uložený do typu " KK("int") ".\n"
      "- Vrací přečtený znak, nebo hodnotu " KK("EOF") " na konci nebo při chybě.\n"
      "\n"
    );
  }

  if (navratNeboSekce("Ukázka - použití fgetc")) return;
  if (pokracovat())
  {
    printh(
      "- Ukázka očekává soubor data/data01.txt vyrobený v předchozí sekci.\n"
      "\n"
      KOD(
      "  char * cesta = \"data/data01.txt\";\n"
      "  FILE * f = fopen(cesta, \"r\");\n"
      "  if (f != NULL)\n"
      "  {\n"
      "    printf(\"  => Obsah souboru %%s:\\n\", cesta);\n"
      "    int znak; // Musí to být int, ne char!\n"
      "    while ( (znak = fgetc(f)) != EOF )\n"
      "    { putchar(znak); }\n"
      "    fclose(f);\n"
      "  }\n"
      "  else\n"
      "  { perror(\"Problem se souborem data01.txt\"); }\n"
      )
      "\n"
    );

    char * cesta = "data/data01.txt";
    FILE * f = fopen(cesta, "r");
    if (f != NULL)
    {
      printf("  => Obsah souboru %s:\n", cesta);
      int znak; // Musí to být int, ne char!
      while ( (znak = fgetc(f)) != EOF )
      { putchar(znak); }
      fclose(f);
    }
    else
    { perror("Problem se souborem data01.txt"); }

    printf("\n");
  }

  if (navratNeboSekce("Čtení formátovaných dat")) return;
  if (pokracovat())
  {
    printh(
      "- Funkce " KK("fscanf") " se chová stejně jako " KK("scanf") ", jen čte ze souboru.\n"
      "\n"
      "- Ukázka očekává soubor data/data02.txt vyrobený v předchozí sekci.\n"
      KOD(
      "...\n"
      "  char slovo[20+1];   int cele;    float desetinne;\n"
      "  while ( fscanf(f, \" %%20[^ \\t;] ; %%d ; %%f \", slovo, &cele, &desetinne) == 3)\n"
      "  {\n"
      "    printf(\"  => slovo: %%s, cele: %%d, desetinne: %%.3f\\n\", slovo, cele, desetinne);\n"
      "  }\n"
      "...\n"
      )
      "\n"
    );
    char * cesta = "data/data02.txt";
    FILE * f = fopen(cesta, "r");
    if (f != NULL)
    {
      printf("  => Obsah souboru %s:\n", cesta);

      char slovo[20+1];   int cele;    float desetinne;
      while (fscanf(f, " %20[^ \t;] ; %d ; %f ", slovo, &cele, &desetinne) == 3)
      {
        printf("  => slovo: %s, cele: %d, desetinne: %.3f\n", slovo, cele, desetinne);
      }

      fclose(f);
    }
    else
    { perror("Problem se souborem data02.txt"); }

    printf("\n");
  }

  if (navratNeboSekce("Čtení ze širokých souborů")) return;
  if (pokracovat())
  {
    printh(
      "- Jde pracovat i se soubory v Unicode kódování.\n"
      "- Každý soubor je potřeba přepnout do širokého módu.\n"
      "  Viz kapitolu Textové řetězce.\n"
      "- Široký mód se u každého souboru či proudu zapíná individuálně.\n"
      "  => Není problém pracovat se širokými soubory, ale na konzoli pracovat\n"
      "     v bajtovém módu.\n"
      "- Ke všem bajtovým funkcím existují široké alternativy, například\n"
      "  " KK("fgetwc") ", " KK("fwscanf") ", " KK("fgetws") ".\n"
      "\n"
      "\n"
    );
  }

  if (navratNeboSekce("Ukázka - použití fwscanf")) return;
  if (pokracovat())
  {
    printh(
      "- Pozor! Funkce " KK("fwscanf") " rozlišuje mezi značkami " KK("%%s") " a " KK("%%ls") ",\n"
      "  " KK("%%[množina") "] a " KK("%%l[množina] ") "pro obyčejné a široké stringy.\n"
      "- Ukázka očekává soubor data/wdata03.txt vyrobený v předchozí sekci.\n"
      "\n"
      KOD(
      "...\n"
      "  if (fwide(f, +1) > 0)\n"
      "  {\n"
      "    wchar_t wslovo[20+1];   int cele;    float desetinne;\n"
      "    while (fwscanf(f, L\" %%20l[^ \\t;] ; %%d ; %%f \", wslovo, &cele, &desetinne) == 3)\n"
      "    {\n"
      "      printf(\"  => slovo: %%ls, cele: %%d, desetinne: %%.3f\\n\", wslovo, cele, desetinne);\n"
      "      // printf také zvládá značku %%ls pro široký string\n"
      "    }\n"
      "  }\n"
      "...\n"
      )
      "\n"
    );

    char * cesta = "data/wdata03.txt";
    FILE * f = fopen(cesta, "r");
    if (f != NULL)
    {
      printf("  => Obsah souboru %s:\n", cesta);
      if (fwide(f, +1) > 0)
      {
        wchar_t wslovo[20+1];   int cele;    float desetinne;
        while (fwscanf(f, L" %20l[^ \t;] ; %d ; %f ", wslovo, &cele, &desetinne) == 3)
        {
          printf("  => slovo: %ls, cele: %d, desetinne: %.3f\n", wslovo, cele, desetinne);
          // printf také zvládá značku %ls pro široký string
        }
      }
      fclose(f);
    }
    else
    { perror("Problem se souborem wdata03.txt"); }

    printf("\n");
  }

  konecKapitoly(__func__);
}



////////////////////////////////////////////////////////////////////////////////
void souboryProCteniIZapis(void)
{
  zacatekKapitoly("Soubory pro čtení i zápis");

  printh(
    "- Soubor musí být otevřený v některém z módů r+, w+, a+.\n"
    "\n"
    "- I v těchto souborech je realizován sekvenční přístup zepředu dozadu,\n"
    "  tj. nejde číst nebo zapisovat pozpátku.\n"
    "- Náhodný přístup není realizován indexováním, ale pomocí funkcí, které\n"
    "  přestaví čtecí/zapisovací hlavu na zadanou pozici v souboru.\n"
    "- Takový přístup jde realizovat i na sekvenčních médiích (pásky), ale za\n"
    "  cenu velké časové složitosti.\n"
    "\n"
    "- Pozor! Do souboru nejde vkládat.\n"
    "- Když se nastaví pozice na začátek nebo doprostřed a něco se tam zapíše,\n"
    "  dojde k přepisu původních bajtů.\n"
    "\n"
    "- Existují i datové proudy pro čtení a zápis - říká se jim duplexní proudy.\n"
    "- Mohou to být například síťové sokety nebo pojmenované roury pro komunikaci\n"
    "  mezi více procesy.\n"
    "- Jsou to však systémově specifické a nekompatibilní řešení. Standard\n"
    "  jazyka C duplexní proudy nezahrnuje.\n"
    // V unixové normě POSIX lze najít funkce popen a pclose, pro vytvoření
    // roury. Je to velice elegantní, protože výsledkem je ukazatel typu FILE*,
    // takže se s tím dá pracovat přesně jako se souborem. Otevírá se to například
    // v módu r+. Ve Windows však toto řešení nebude fungovat.
    // Windows mají funkce CreateNamedPipe a ConnectNamedPipe, které ale
    // nevyužívají standardní souborové rozhraní.
    "\n"
  );

  if (navratNeboSekce("Funkce fseek")) return;
  if (pokracovat())
  {
    printh(
      "- Tato funkce nastaví hlavu na zadanou pozici v souboru.\n"
      "\n"
      KOD(
      "int fseek(FILE *soubor, long offset, int odkud); \n"
      )
      "\n"
      "" KK("offset") "\n"
      "- Posun v bajtech vůči pozici specifikované parametrem " KK("odkud") ".\n"
      "- Může to být kladné i záporné číslo.\n"
      "\n"
      "" KK("odkud") "\n"
      "- Základní pozice pro posunutí " KK("offset") ".\n"
      "  " KK("SEEK_SET") " - Výchozí pozice je začátek souboru.\n"
      "  " KK("SEEK_CUR") " - Výchozí pozice je aktuální pozice v souboru.\n"
      "  " KK("SEEK_END") " - Výchozí pozice je konec souboru.\n"
      "\n"
      "- Funkce vrací 0 v případě úspěchu.\n"
      "- Funkce vrací -1 v případě neúspěchu a pak nastaví chybový kód do " KK("errno") ".\n"
      "\n"
      "- Funkce zároveň maže příznak dosažení konce souboru.\n"
      "- Nepoužívat po " KK("ungetc") " - ruší její efekt.\n"
      "\n"
    );
  }

  if (navratNeboSekce("Další funkce")) return;
  if (pokracovat())
  {
    printh(
      "" KK("ftell") "\n"
      "- Vrací aktuální pozici hlavy od začáku souboru v bajtech.\n"
      "\n"
      "" KK("rewind") "\n"
      "- Nastaví pozici hlavy na začátek souboru.\n"
      "\n"
      "" KK("fgetpos") "\n"
      "- Vrací aktuální pozici hlavy od začátku - alternativa " KK("ftell") ".\n"
      "\n"
      "" KK("fsetpos") "\n"
      "- Nastaví pozici hlavy od začátku - alternativa " KK("fseek") " + " KK("SEEK_SET") ".\n"
      "\n"
      "\n"
    );
  }

  if (navratNeboSekce("Ukázka - zápis, čtení a skákání v souboru")) return;
  if (pokracovat())
  {
    printh(
      "- Ukázka používá soubor data/data00.txt\n"
      "\n"
      KOD(
      "  FILE * soubor = fopen(\"data/data00.txt\", \"r+\");\n"
      "  if (soubor != NULL)\n"
      "  {\n"
      "    fseek(soubor, 0L, SEEK_END); // skok na konec - a+ by to zařídil taky\n"
      "    fprintf(soubor, \"\\nPejsku nas,\\nco delas,\\nzes tak vesel stale.\\n\");\n"
      "    fprintf(soubor, \"Rek bych vam,\\nnevim sam.\\nHop, a skakal dale.\\n\");\n"
      "\n"
      "    printf(\" => Posledni radek\\n\");\n"
      "    // Skok o řádek zpět\n"
      "    fseek(soubor, -strlen(\"Hop, a skakal dale.\\n\"), SEEK_END);\n"
      "    zpracujRadek(soubor); // Přečte jeden řádek textu\n"
      "\n"
      "    printf(\" => Prvni radek\\n\");\n"
      "    rewind(soubor);\n"
      "    zpracujRadek(soubor);\n"
      "\n"
      "    fclose(soubor);\n"
      "  }\n"
      "  else\n"
      "  { perror(\"Chyba pri otevirani souboru\"); }\n"
      )
      "\n"
    );

    FILE * soubor = fopen("data/data00.txt", "r+");
    if (soubor != NULL)
    {
      fseek(soubor, 0L, SEEK_END); // skok na konec - a+ by to zařídil taky
      fprintf(soubor, "\nPejsku nas,\nco delas,\nzes tak vesel stale.\n");
      fprintf(soubor, "Rek bych vam,\nnevim sam.\nHop, a skakal dale.\n");

      printf(" => Posledni radek\n");
      // Skok o řádek zpět
      fseek(soubor, -strlen("Hop, a skakal dale.\n"), SEEK_END);
      zpracujRadek(soubor); // Přečte jeden řádek textu

      printf(" => Prvni radek\n");
      rewind(soubor);
      zpracujRadek(soubor);

      fclose(soubor);
    }
    else
    { perror("Chyba pri otevirani souboru"); }

    printf("\n");
  }

  konecKapitoly(__func__);
}



////////////////////////////////////////////////////////////////////////////////
void binarniSoubory(void)
{
  zacatekKapitoly("Zpracování binárních souborů");

  printh(
    "- V unixu se dá s každým souborem zacházet jako s binárním.\n"
    "- Ve Windows se rozlišují binární a textové soubory kvůli dvojznakovému\n"
    "  konci řádku. Viz první kapitola tohoto dílu.\n"
    "- Ve Windows se ve funkci " KK("fopen") " k řetězci pro výběr módu přidává\n"
    "  přípona b, například rb, wb, rb+ atd.\n"
    "- Tato přípona se může přidat i v unixu, ale bude ignorována.\n"
    "\n"
    "- Binární soubor se stává binárním tím, jak se s ním pracuje.\n"
    "- Čtení a zápis po bajtech se dělá stejně jako v textovém souboru, pomocí\n"
    "  funkcí " KK("fgetc") " a " KK("fputc") ".\n"
    "- Přečtené hodnoty se ale budou interpretovat jako bajty, ne znaky.\n"
    "\n"
    "- Obvyklejší je čtení a zápis po blocích bajtů zadané délky.\n"
    "- Tyto bloky se pak následně pomocí operátoru přetypování převedou na\n"
    "  hodnoty vhodných datových typů.\n"
    "\n"
  );

  if (navratNeboSekce("Pevné číselné typy")) return;
  if (pokracovat())
  {
//    int8_t bajt;         int16_t kratke;
//    int32_t normalni;    int64_t dlouhe;

    printh(
      "- Pro přetypování celočíselných hodnot je kvůli přenositelnosti lepší\n"
      "  se vyhnout standardním typům jako " KK("int") ", " KK("short") ", či " KK("long") ".\n"
      "- Tyto typy mají na různých architekturách procesorů různé velikosti.\n"
      "\n"
      "- Modul stdint.h definuje číselné typy s pevnou velikostí a pomocné\n"
      "  konstanty (minimum, maximum atd.).\n"
      "\n"
      KOD(
      "  int8_t bajt;         int16_t kratke;\n"
      "  int32_t normalni;    int64_t dlouhe;\n"
      )
      "\n"
      "  => sizeof(int8_t):  %ld bajt  = %ld bitů\n"
      "  => sizeof(int16_t): %ld bajty = %ld bitů\n"
      "  => sizeof(int32_t): %ld bajty = %ld bitů\n"
      "  => sizeof(int64_t): %ld bajtů = %ld bitů\n"
      "\n"
      , sizeof(int8_t), sizeof(int8_t)*8
      , sizeof(int16_t), sizeof(int16_t)*8
      , sizeof(int32_t), sizeof(int32_t)*8
      , sizeof(int64_t), sizeof(int64_t)*8
    );
    printh(
      "- Racionální typy " KK("float") ", " KK("double") " a " KK("long double") " mají pevnou\n"
      "  velikost danou normou IEEE 754.\n"
      "\n"
    );
  }

  if (navratNeboSekce("Endianita")) return;
  if (pokracovat())
  {
    printh(
      "- Je to pořadí ukládání a zpracování bajtů u vícebajtových hodnot počítačem.\n"
      "- Někdy se tomu říká problém \"little\" a \"big endian\".\n"
      "\n"
      PODSEKCE("Little endian")
      "- Tento systém ukládá bajty v pořadí od nejméně významného po nejvýznamnější.\n"
      "- Je to opačně, než jak zapisujeme čísla na papíře:\n"
      KOD(
      "\n"
      "  hodnota: 0x00FFA13D\n"
      "\n"
      "  adresa: 00 01 02 03\n"
      "         +--+--+--+--+\n"
      "  paměť: |3D|A1|FF|00|\n"
      "         +--+--+--+--+\n"
      )
      "\n"
      "- Tento systém používá například architektura Intel x86.\n"
      "- Výhodnou je, že přidáváním bajtů se nemění adresa, takže když se bajt\n"
      "  přetypuje na int, přidávají se nulové bajty doprava a původní hodnota\n"
      "  i adresa začátku čísla zůstávají stejné.\n"
      "\n"
      PODSEKCE("Big endian")
      "- Tento systém ukládá bajty v pořadí, jak je zapisujeme na papíře, tj.\n"
      "  čím vyšší řád, tím více vlevo.\n"
      "- Tento systém používá architektura SPARC.\n"
      "\n"
      "\n"
      "=> Problém s kompatibilitou těchto systémů se řeší ukládáním dat v textové\n"
      "   podobě (např. LibreOffice), nebo prací po jednotlivých bajtech.\n"
      "\n"
    );
  }

  if (navratNeboSekce("Zápis po blocích - fwrite")) return;
  if (pokracovat())
  {
    printh(
      "\n"
      KOD(
      "size_t fwrite(const void *ptr, size_t size, size_t nmemb, FILE *stream);\n"
      )
      "\n"
      "- Funkce " KK("fwrite") " zapisuje binární data do souboru otevřeného pro zápis.\n"
      "\n"
      "" KK("ptr") "    - ukazatel na libovolná data (typicky pole)\n"
      "" KK("size") "   - počet bajtů které zabírá jeden prvek\n"
      "" KK("nmemb") "  - počet prvků pole, který se bude zapisovat\n"
      "" KK("stream") " - soubor nebo proud otevřený pro (binární) zápis\n"
      "\n"
      "- Funkce vrací počet zapsaných prvků o velikosti " KK("size") " bajtů.\n"
      "\n"
    );
  }

  if (navratNeboSekce("Ukázka - použití fwrite")) return;
  if (pokracovat())
  {
    printh(
      "\n"
      KOD(
      "  char * cesta = \"data/data04.bin\";\n"
      "  FILE * f = fopen(cesta, \"wb\"); // b jen kvůli Windows\n"
      "  if (f != NULL)\n"
      "  {\n"
      "    int64_t pole[] = {11111, -22222, 33333, -44444, 55555};\n"
      "    size_t n = fwrite(pole, sizeof(int64_t), 5, f);\n"
      "\n"
      "    fclose(f);\n"
      "    printg(\"  => Podívej se do souboru %%s.\\n\", cesta);\n"
      "    printg(\"  => n: %%ld\\n\", n);\n"
      "  }\n"
      "  else\n"
      "  { perror(\"Problem se souborem data04.bin\"); }\n"
      )
      "\n"
    );

    char * cesta = "data/data04.bin";
    FILE * f = fopen(cesta, "wb"); // b jen kvůli Windows
    if (f != NULL)
    {
      int64_t pole[] = {11111, -22222, 33333, -44444, 55555};
      size_t n = fwrite(pole, sizeof(int64_t), 5, f);

      fclose(f);
      printg("  => Podívej se do souboru %s.\n", cesta);
      printg("  => n: %ld\n", n);
    }
    else
    { perror("Problem se souborem data04.bin"); }

    printf("\n");
  }

  if (navratNeboSekce("Čtení po blocích - fread")) return;
  if (pokracovat())
  {
    printh(
      "\n"
      KOD(
      "size_t fread(void *ptr, size_t size, size_t nmemb, FILE *stream);\n"
      )
      "\n"
      "- Funkce " KK("fread") " čte binární data ze souboru otevřeného pro čtení.\n"
      "\n"
      "" KK("ptr") "    - ukazatel na libovolná data (typicky pole)\n"
      "" KK("size") "   - počet bajtů, které zabírá jeden prvek\n"
      "" KK("nmemb") "  - počet prvků, který se má číst\n"
      "" KK("stream") " - soubor nebo proud otevřený pro (binární) čtení\n"
      "\n"
      "- Funkce vrací počet přečtených prvků o velikosti " KK("size") " bajtů.\n"
      "\n"
    );
  }

  if (navratNeboSekce("Ukázka - použití fwread")) return;
  if (pokracovat())
  {
    printh(
      "- Ukázka čte ze souboru data/data04.bin z předchozího příkladu.\n"
      "\n"
      KOD(
      "  char * cesta = \"data/data04.bin\";\n"
      "  FILE * f = fopen(cesta, \"rb\"); // b jen kvůli Windows\n"
      "  if (f != NULL)\n"
      "  {\n"
      "    int64_t pole[5];\n"
      "    size_t n = fread(pole, sizeof(int64_t), 5, f);\n"
      "    fclose(f);\n"
      "\n"
      "    printf(\"  => pole: { %%ld\", pole[0]);\n"
      "    for (unsigned int i = 1; i < n; ++i) { printf(\", %%ld\", pole[i]); }\n"
      "    printf(\" }\\n\");\n"
      "\n"
      "    printg(\"  => n: %%ld\\n\", n);\n"
      "  }\n"
      "  else\n"
      "  { perror(\"Problem se souborem data04.bin\"); }\n"
      )
      "\n"
    );

    char * cesta = "data/data04.bin";
    FILE * f = fopen(cesta, "rb"); // b jen kvůli Windows
    if (f != NULL)
    {
      int64_t pole[5];
      size_t n = fread(pole, sizeof(int64_t), 5, f);
      fclose(f);

      printf("  => pole: { %ld", pole[0]);
      for (unsigned int i = 1; i < n; ++i) { printf(", %ld", pole[i]); }
      printf(" }\n");

      printg("  => n: %ld\n", n);
    }
    else
    { perror("Problem se souborem data04.bin"); }

    printf("\n");
  }

  konecKapitoly(__func__);
}

////////////////////////////////////////////////////////////////////////////////

#define TZ_STR_LEN (8+1)
typedef struct zaznam
{
  char str[TZ_STR_LEN];
  int32_t num;
  double dbl;
} Tzaznam;

/** Uloží položky z proměnné typu Tzaznam do pole bajtů.
 *
 * \param z const Tzaznam* Ukazatel na záznam, který se zapakuje do pole.
 * \param pole void* Pole bajtů, kam se zapíší data ze struktury. Musí být dost dlouhé.
 * \return void* Vrací ukazatel na zapisované pole bajtů. To se hodí pro použití
 *               ve výrazech.
 */
void * zapakuj(const Tzaznam * z, void *pole)
{
  char * p = (char *)pole;

  int offset = 0;
  memcpy(p, z->str, TZ_STR_LEN);
  offset += TZ_STR_LEN;
  memcpy(p+offset, &(z->num), sizeof(z->num));
  offset += sizeof(z->num);
  memcpy(p+offset, &(z->dbl), sizeof(z->dbl));

  // p+offset je tady totéž jako &p[offset]

  return p;
}


/** Rozbalí data z pole bajtů do proměnné typu Tzaznam.
 *
 * \param pole const void* Ukazatel na pole se zapakovanými daty.
 * \param z Tzaznam* Ukazatel na proměnnou, kam se data z pole zapíší.
 * \return Tzaznam* Vrací ukazatel na zapisovanou proměnnou.
 */
Tzaznam * rozpakuj(const void *pole, Tzaznam *z)
{
  const char * p = (const char *)pole;

  int offset = 0;
  memcpy(z->str, p, TZ_STR_LEN);
  offset += TZ_STR_LEN;
  memcpy(&(z->num), p+offset, sizeof(z->num));
  offset += sizeof(z->num);
  memcpy(&(z->dbl), p+offset, sizeof(z->dbl));

  // p+offset je tady totéž jako &p[offset]

  return z;
}


void binarniStruktury(void)
{
  zacatekKapitoly("Struktury v binárních souborech");

  printh(
    "- Binárně se dají ukládat i struktury, ale není to bezpečné.\n"
    "- Položky uvnitř struktur bývají zarovnány na adresy dělitelné 2 nebo 4.\n"
    "- Mezi položky může překladač vkládat volné místo.\n"
    "- Překladače mohou na různých architekturách vyrobit vnitřní uspořádání\n"
    "  stejné struktury rozdílně.\n"
    "  => Vytváří to další problémy s přenositelností takových dat.\n"
    "\n"
    "- Mnohem bezpečnější je ukládat prvky samostatně po složkách, ovšem za cenu\n"
    "  horší časové složitosti programu.\n"
    "- Překladače nabízejí také možnost vyrábět tzv. pakované struktury, bez\n"
    "  mezer mezi položkami. Není to ale součástí standardu jazyka C a některé\n"
    "  architektury přímo zakazují, aby data začínala na lichých adresách.\n"
    "\n"
    "\n"
  );

  if (navratNeboSekce("Ukázka ukládání pakovaných struktur")) return;
  if (pokracovat())
  {
    printh(
      "- Celý kód funkce " KK("zapakuj") " najdeš ve zdrojovém textu.\n"
      "\n"
      KOD(
      "  #define NN 3\n"
      "  Tzaznam pole[NN] = {\n"
      "    { .str=\"Brno\", .num = 13, .dbl=2.2 },\n"
      "    { .str=\"Karvina\", .num = 17, .dbl=8.8 },\n"
      "    { .str=\"Hrcava\", .num = 19, .dbl=17.17 },\n"
      "  };\n"
      "\n"
      "  // velikost pakovaného záznamu\n"
      "  int pvelikost = TZ_STR_LEN+sizeof(int32_t)+sizeof(double);\n"
      "  char ppole[NN*pvelikost];\n"
      "\n"
      "  for (int i = 0; i < NN; ++i) { zapakuj(&pole[i], &ppole[i*pvelikost]); }\n"
      "\n"
      "  size_t n = fwrite(ppole, pvelikost, NN, f);\n"
      )
      "\n"
    );

    char * cesta = "data/data05.bin";
    FILE * f = fopen(cesta, "wb"); // b jen kvůli Windows
    if (f != NULL)
    {
      #define NN 3
      Tzaznam pole[NN] = {
        { .str="Brno", .num = 13, .dbl=2.2 },
        { .str="Karvina", .num = 17, .dbl=8.8 },
        { .str="Hrcava", .num = 19, .dbl=17.17 },
      };

      // velikost pakovaného záznamu
      int pvelikost = TZ_STR_LEN+sizeof(int32_t)+sizeof(double);
      char ppole[NN*pvelikost];

      for (int i = 0; i < NN; ++i) { zapakuj(&pole[i], &ppole[i*pvelikost]); }

      size_t n = fwrite(ppole, pvelikost, NN, f);
      fclose(f);

      printg("  => Podívej se do souboru %s.\n", cesta);
      printg("  => n: %ld\n", n);
      printg("  => pvelikost: %d\n", pvelikost);
      printg("  => sizeof(Tzaznam): %ld\n", sizeof(Tzaznam));
    }
    else
    { perror("Problem se souborem data05.bin"); }

    printf("\n");
  }

  if (navratNeboSekce("Ukázka čtení pakovaných struktur")) return;
  if (pokracovat())
  {
    printh(
      "- Celý kód funkce rozpakuj najdeš ve zdrojovém textu.\n"
      "- \n"
      KOD(
      "  #define NN 3\n"
      "  Tzaznam pole[NN];\n"
      "\n"
      "  // velikost pakovaného záznamu\n"
      "  int pvelikost = TZ_STR_LEN+sizeof(int32_t)+sizeof(double);\n"
      "  char ppole[NN*pvelikost];\n"
      "  size_t n = fread(ppole, pvelikost, NN, f);\n"
      "  fclose(f);\n"
      "\n"
      "  for (unsigned int i = 0; i < n; ++i) { rozpakuj(&ppole[i*pvelikost], &pole[i]); }\n"
      )
      "\n"
    );

    char * cesta = "data/data05.bin";
    FILE * f = fopen(cesta, "rb"); // b jen kvůli Windows
    if (f != NULL)
    {
      #define NN 3
      Tzaznam pole[NN];

      // velikost pakovaného záznamu
      int pvelikost = TZ_STR_LEN+sizeof(int32_t)+sizeof(double);
      char ppole[NN*pvelikost];
      size_t n = fread(ppole, pvelikost, NN, f);
      fclose(f);

      for (unsigned int i = 0; i < n; ++i) { rozpakuj(&ppole[i*pvelikost], &pole[i]); }
      printf("  => pole: {");
      for (unsigned int i = 0; i < n; ++i)
      {
        printf("{%s, %d, %g}, ", pole[i].str, pole[i].num, pole[i].dbl);
      }
      printf("}\n");
      printg("  => n: %ld\n", n);
      printg("  => pvelikost: %d\n", pvelikost);
      printg("  => sizeof(Tzaznam): %ld\n", sizeof(Tzaznam));
    }
    else
    { perror("Problem se souborem data05.bin"); }

    printf("\n");
  }

  konecKapitoly(__func__);
}

